#include <sys/types.h>
#include <dlfcn.h>
#include <string>
#include <vector>
#include <cctype>
#include <cstring>
#include <cmath>
#include <map>
#include <set>

#include <arpa/telnet.h>

#include "coroutine.h"

static struct LibcGet
{
    void *f;
    LibcGet() : f(dlopen("libc.so.6",RTLD_NOW)) {}
    ~LibcGet() { dlclose(f); }
    void *operator() (const char *s) { return dlsym(f, s); }  
} caller;

static const std::string UCase(const std::string& s)
{
    std::string result(s);
    for(unsigned a=0; a<result.size(); ++a)
        result[a] = std::toupper(result[a]);
    return result;
}
static bool is_valid_tagname_char(const char c)
{
    return  (c >= 'A' && c <= 'Z')
         || (c >= 'a' && c <= 'z')
         || (c >= '0' && c <= '9')
         || (c == '-' || c == '_' || c == '!');
}

template<typename S>
static const std::set<S> ExplodeSet(const std::string& sep, const S& subj)
{
    std::set<S> result;
    typename S::size_type pos=0;
    while(pos < subj.size())
    {
        typename S::size_type found = subj.find(sep, pos);
        if(found == subj.npos) break;
        result.insert(subj.substr(pos, found-pos));
        pos = found + sep.size();
    }
    if(pos < subj.size()) result.insert(subj.substr(pos));
    return result;
}
template<typename S>
static const S JoinSet(const std::string& sep, const std::set<S>& subj)
{
    S result;
    bool first=true;
    for(typename std::set<S>::const_iterator i = subj.begin(); i != subj.end(); ++i)
    {
        if(first)first=false; else result += sep;
        result += *i;
    }
    return result;
}

struct ColorInfo { const char*name; unsigned color; };
static bool FindColorInfo(const ColorInfo&a, const ColorInfo& b)
{
    return std::strcmp(a.name, b.name) < 0;
}

static unsigned TranslateColorNameToRGB(const std::string& name)
{
    #define RGB(r,g,b) (((r)<<16) + ((g)<<8) + ((b)))
    std::string nameu = UCase(name);
    if(nameu.size() == 4 && nameu[0] == '#')
    {
        unsigned val = strtol(nameu.c_str()+1, 0, 16);
        unsigned r = (val&0xF00)>>8;  r*=0x11;
        unsigned g = (val&0xF0)>>4;   g*=0x11;
        unsigned b = (val&0xF)>>0;    b*=0x11;
        return RGB(r,g,b);
    }
    if(nameu.size()==7 && nameu[0] == '#')
    {
        return std::strtol(nameu.c_str()+1, 0, 16);
    }
    static const ColorInfo colors[] =
    {
#include "colors.inc"
    };
    
    ColorInfo dummy; dummy.name = nameu.c_str();
    const ColorInfo* endcolors = colors+ (sizeof(colors)/sizeof(*colors));
    
    const ColorInfo* color = std::lower_bound( colors, endcolors, dummy, FindColorInfo);
    if(color != endcolors)
        return color->color;
    return 0xC0C0C0;
    #undef RGB
}

static double RgbDiff(unsigned rgb1, unsigned rgb2)
{
    return std::pow( (int)((rgb1>>16)&0xFF) - (int)((rgb2>>16)&0xFF), 2.0)
         + std::pow( (int)((rgb1>> 8)&0xFF) - (int)((rgb2>> 8)&0xFF), 2.0)
         + std::pow( (int)((rgb1>> 0)&0xFF) - (int)((rgb2>> 0)&0xFF), 2.0);
}

static unsigned CreateXtermColor(const unsigned rgb)
{
    static const unsigned char sixtable[6] = {0x00,0x5F,0x87,0xAF,0xD7,0xFF};
    #define graytable(n) (8 + 10*(n))  /* n=0..23 */
    unsigned r = (rgb>>16)&0xFF, g = (rgb>>8)&0xFF, b = (rgb>>0)&0xFF;
    
    // Try six-algorithm
    unsigned r6 = r*6/256, g6=g*6/256, b6=b*6/256;
    unsigned six_result = (sixtable[r6]<<16) + (sixtable[g6]<<8) + (sixtable[b6]);
    
    // Try grayscale-algorithm
    unsigned gray_value  = (r<8) ? 0 : (r>0xEE ? 23 : (r-8)/10);
    unsigned gray_result = graytable(gray_value);
    
    if(RgbDiff(gray_result,rgb) < RgbDiff(six_result,rgb))
    {
        return 232 + gray_value;
    }
    return 16 + b6 + g6*6 + r6*6*6;
}

class Transceiver
{
private:
    std::string PendingResult;
    std::string PendingMacro;
    enum {
        DefaultMode,
        SecureMode,
        LockedMode,
        OpenMode
    } CurrentMode;
    enum {
        ModeLocked,
        UntilNl,
        UntilTagEnd
    } ModeMode;
    bool ClosingTagsMode;
    
    struct FontDef
    {
        std::string fontname;
        std::string fontsize;
        std::string forecolordefs;
        std::string backcolordefs;
        
        const std::string str() const
        {
            return "name(" + fontname + ")"
                 + "size(" + fontsize + ")"
                 + "fc(" + forecolordefs + ")"
                 + "bc(" + backcolordefs + ")";
        }
    };
    struct Formatting
    {
        unsigned B, I, U, S, Bright;
        std::vector<FontDef> FontDefs;
        
        void Reset()
        {
            B=I=U=S=Bright=0;
            FontDefs.clear();
        }
        void Add(unsigned& var, int value)
        {
            long result = var; result += value;
            if(result < 0) result = 0;
            if(result > 31) result = 31;
            var = result;
        }
        void PopFontDef()
        {
            //fprintf(stderr, "Popped FontDef\n");
            if(!FontDefs.empty()) FontDefs.resize(FontDefs.size()-1);
        }
        void PushFontDef(const FontDef& def)
        {
            //fprintf(stderr, "Added FontDef: %s\n", def.str().c_str());
            if(FontDefs.size() > 31) FontDefs.erase(FontDefs.begin(), FontDefs.begin()+10);
            FontDefs.push_back(def);
        }
        
        Formatting() { Reset(); }
    } Formatting;
    
    struct Tag
    {
        std::string TagName;
        bool IsClosingTag;
        
        // name and value can both be empty
        std::vector<std::pair<std::string/*name*/, std::string/*value*/> > Params;
    };
    struct TagDef
    {
        std::string MacroExpansion;
        std::vector<std::pair<std::string/*name*/, std::string/*defaultvalue*/> > AttrList;
        int Tag;
        std::string Flags; //?
        bool Is_OpenTag;  // Can user redefine this?
        bool Is_EmptyTag; // Does not require closing tag?
        
        void (Transceiver::*Handler) (const struct TagDef&, const struct Tag&);
        
        TagDef(): Tag(-1),Is_OpenTag(false),Is_EmptyTag(true) { }
    };
    std::map<std::string/*tagname*/, TagDef> TagDefs;
    struct EntityDef
    {
        std::string Value;
        std::string Desc;
        bool Is_PubEnt;  // Can user query this?
        bool Is_OpenEnt; // Can user redefine this?
        bool Is_Recursive;
        
        EntityDef(): Is_PubEnt(false),Is_OpenEnt(false), Is_Recursive(false) { }
    };
    std::map<std::string, EntityDef> EntityDefs;
    
private:
    struct AttrMatchResult
    {
        std::map<std::string, std::string> matched;
        std::vector<std::string> unmatched;
    };
    AttrMatchResult DoAttrMatching(const TagDef& def, const Tag& tag)
    {
        AttrMatchResult result;
        /* Set default values first. */
        for(unsigned a=0; a<def.AttrList.size(); ++a)
        {
            const std::pair<std::string,std::string>& attr = def.AttrList[a];
            if(!attr.second.empty())
                result.matched[attr.first] = attr.second;
        }
        /* Now check what was given. */
        for(unsigned a=0; a<tag.Params.size(); ++a)
        {
            /* "a"="b" */
            const std::string& key   = tag.Params[a].first;
            const std::string& value = tag.Params[a].second;
            
            if(!key.empty())
            {
                result.matched[key] = value;
                //fprintf(stderr, "Adding '%s' as '%s'\n", value.c_str(), key.c_str());
            }
            else
            {
                /* just "a" */
                /* It can be a positional attribute. */
                /* If "value" is empty, use the default value instead. */
                if(value.empty()) continue;
                if(a < def.AttrList.size())
                {
                    result.matched[def.AttrList[a].first] = value;
                    //fprintf(stderr, "Adding '%s' as '%s' (posi)\n",
                    //    value.c_str(), def.AttrList[a].first.c_str());
                }
                else
                {
                    result.unmatched.push_back(value);
                    //fprintf(stderr, "Adding '%s' as raw\n",value.c_str());
                }
            }
        }
        return result;
    }
    
    const std::vector<std::pair<std::string/*name*/, std::string/*value*/> >
        ParseAttrList(const std::string& s,
                      bool default_is_name)
    {
        std::vector<std::pair<std::string/*name*/, std::string/*value*/> > result;
        
        std::vector<std::string> tokens;
        std::string::size_type tagbegin=0;
        bool tagmode=false;
        char quotemode=0;
        for(std::string::size_type a=0; a<s.size(); ++a)
        {
            char ch = s[a];
            if(!quotemode && std::isspace(ch))
                { if(tagmode) tokens.push_back(s.substr(tagbegin,a-tagbegin));
                  tagmode=false; continue; }
            if(!tagmode) { tagmode=true; tagbegin=a; }
            if(quotemode && ch==quotemode) { quotemode=false; continue; }
            if(!quotemode && (ch=='"' || ch=='\'')) { quotemode=ch; continue; }
        }
        if(tagmode) tokens.push_back(s.substr(tagbegin));
        for(std::string::size_type a=0; a<tokens.size(); ++a)
        {
            const std::string& token = tokens[a];
            
            //fprintf(stderr, "token: '%s'\n", token.c_str());
            
            std::string::size_type eqpos = token.find('=');
            std::string::size_type q1pos = token.find('\'');
            std::string::size_type q2pos = token.find('"');
            if(eqpos != token.npos)
            {
                if(q1pos != token.npos && q1pos < eqpos) eqpos=token.npos;
                if(q2pos != token.npos && q2pos < eqpos) eqpos=token.npos;
            }
            if(eqpos != token.npos)
                result.push_back(std::make_pair(token.substr(0, eqpos),
                                                token.substr(eqpos+1)));
            else if(default_is_name)
                result.push_back(std::make_pair(token, ""));
            else
                result.push_back(std::make_pair("", token));
        }
        return result;
    }
    void ParseAttrList(TagDef& result, const std::string& attrlist)
    {
        result.AttrList = ParseAttrList(attrlist, true);
    }
    
    void DefaultTagHandler(const TagDef& def, const Tag& tag)
    {
        // This should just do the macro expansion.
        // However, if this is a closing tag, it should also
        // do the closing versions of the tags included in the macro.
        // For this, we use the esc[2;0z and esc[2;1z macroes.
        
        std::string macro = def.MacroExpansion;
        if(tag.IsClosingTag)
            macro = "\33[2;1z" + macro + "\33[2;0z";
        DoMacroExpansion(macro);
    }

    void HandleElementMetaTag(const TagDef& def, const Tag& tag)
    {
        AttrMatchResult attrs = DoAttrMatching(def, tag);
        std::string elname = UCase(attrs.matched["ELEMENT-NAME"]);
        {std::map<std::string, TagDef>::iterator i = TagDefs.find(elname);
        if(i != TagDefs.end())
        {
            /* Check if we can redefine the tag */
            if(!i->second.Is_OpenTag && CurrentMode != SecureMode) return; // ignore!
        }}
        TagDef& result = TagDefs[elname];
        result.Handler = &Transceiver::DefaultTagHandler;
        result.MacroExpansion = attrs.matched["DEFINITION"];
        result.Tag            = std::atoi(attrs.matched["TAG"].c_str());
        result.Flags          = attrs.matched["FLAG"];
        ParseAttrList(result, attrs.matched["ATT"]);
        result.Is_EmptyTag = false;
        result.Is_OpenTag  = false;
        for(unsigned a=0; a<attrs.unmatched.size(); ++a)
        {
            const std::string param = UCase(attrs.unmatched[a]);
            if(param == "OPEN") result.Is_OpenTag = true;
            if(param == "EMPTY") result.Is_EmptyTag = true;
            if(param == "DELETE") { TagDefs.erase(elname); return; }
        }
    }
    void HandleAttListMetaTag(const TagDef& def, const Tag& tag)
    {
        AttrMatchResult attrs = DoAttrMatching(def, tag);
        std::string elname = UCase(attrs.matched["ELEMENT-NAME"]);
        std::map<std::string, TagDef>::iterator i = TagDefs.find(elname);
        if(i == TagDefs.end()) return;
        /* Check if we can redefine the tag */
        if(!i->second.Is_OpenTag && CurrentMode != SecureMode) return; // ignore!
        ParseAttrList(i->second, attrs.matched["ATTRIBUTE-LIST"]);
    }
    void HandleEntityMetaTag(const TagDef& def, const Tag& tag)
    {
        AttrMatchResult attrs = DoAttrMatching(def, tag);
        std::string enname = attrs.matched["NAME"];
        {std::map<std::string, EntityDef>::iterator i = EntityDefs.find(enname);
        if(i != EntityDefs.end())
        {
            /* Check if we can redefine the tag */
            if(!i->second.Is_OpenEnt && CurrentMode != SecureMode) return; // ignore!
        }}
        EntityDef& result = EntityDefs[enname];

        const std::string& newvalue = attrs.matched["VALUE"];
        
        enum { Replace, Delete, Add, Remove } mode = Replace;
        result.Desc  = attrs.matched["DESC"];
        result.Is_Recursive = true;
        result.Is_OpenEnt = true;
        for(unsigned a=0; a<attrs.unmatched.size(); ++a)
        {
            const std::string param = UCase(attrs.unmatched[a]);
            if(param == "PRIVATE") result.Is_PubEnt = result.Is_OpenEnt = false;
            if(param == "PUBLISH") result.Is_PubEnt = true;
            if(param == "DELETE") mode = Delete;
            if(param == "ADD")    mode = Add;
            if(param == "REMOVE") mode = Remove;
        }
        switch(mode)
        {
            case Replace: result.Value = newvalue; break;
            case Delete:  { EntityDefs.erase(enname); return; }
            case Add: case Remove:
                std::set<std::string> curvalues = ExplodeSet("|", result.Value);
                std::set<std::string> newvalues = ExplodeSet("|", newvalue);
                // remove.
                for(std::set<std::string>::const_iterator
                    i = newvalues.begin(); i != newvalues.end(); ++i)
                        curvalues.erase(*i);
                
                // remove + add.
                if(mode == Add)
                {
                    for(std::set<std::string>::const_iterator
                        i = newvalues.begin(); i != newvalues.end(); ++i)
                            curvalues.insert(*i);
                }
                result.Value = JoinSet("|", curvalues);
                break;
        }
    }
    
    void HandleBoldTag(const TagDef& def, const Tag& tag)
        { Formatting.Add(Formatting.B, tag.IsClosingTag?-1:1); RedoFormatting(); }
    void HandleItalicTag(const TagDef& def, const Tag& tag)
        { Formatting.Add(Formatting.I, tag.IsClosingTag?-1:1); RedoFormatting(); }
    void HandleUnderlineTag(const TagDef& def, const Tag& tag)
        { Formatting.Add(Formatting.U, tag.IsClosingTag?-1:1); RedoFormatting(); }
    void HandleStrikeoutTag(const TagDef& def, const Tag& tag)
        { Formatting.Add(Formatting.S, tag.IsClosingTag?-1:1); RedoFormatting(); }
    void HandleColorTag(const TagDef& def, const Tag& tag)
    {
        if(tag.IsClosingTag) { Formatting.PopFontDef(); RedoFormatting(); return; }
        FontDef fnt;
        AttrMatchResult attrs = DoAttrMatching(def, tag);
        fnt.forecolordefs = attrs.matched["FORE"];
        fnt.backcolordefs = attrs.matched["BACK"];
        Formatting.PushFontDef(fnt);
        RedoFormatting();
    }
    void HandleBrightTag(const TagDef& def, const Tag& tag)
        { Formatting.Add(Formatting.Bright, tag.IsClosingTag?-1:1); RedoFormatting(); }
    void HandleFontTag(const TagDef& def, const Tag& tag)
    {
        if(tag.IsClosingTag) { Formatting.PopFontDef(); RedoFormatting(); return; }
        FontDef fnt;
        AttrMatchResult attrs = DoAttrMatching(def, tag);
        fnt.fontname = attrs.matched["FACE"];
        fnt.fontsize = attrs.matched["SIZE"];
        fnt.forecolordefs = attrs.matched["COLOR"];
        fnt.backcolordefs = attrs.matched["BACK"];
        Formatting.PushFontDef(fnt);
        RedoFormatting();
    }
    
    
    void HandleNobrTag(const TagDef& def, const Tag& tag) { }
    void HandleParaTag(const TagDef& def, const Tag& tag) { }
    void HandleBRTag(const TagDef& def, const Tag& tag) { }
    void HandleSBRTag(const TagDef& def, const Tag& tag) { }

    void HandleSupportTag(const TagDef& def, const Tag& tag)
    {
        /* Support tag is not supported :P */
    }

    void InitDefaultTags()
    {
        #define DefAt(t, name, defval) t.AttrList.push_back(std::make_pair(name,defval))
        
        /* COMMAND TAGS */
    
        {TagDef ElementMetaTag;
            DefAt(ElementMetaTag, "ELEMENT-NAME","");
            DefAt(ElementMetaTag, "DEFINITION",  "");
            DefAt(ElementMetaTag, "ATT",         "");
            DefAt(ElementMetaTag, "TAG",         "");
            DefAt(ElementMetaTag, "FLAG",        "");
            ElementMetaTag.Handler = &Transceiver::HandleElementMetaTag;
        TagDefs["!ELEMENT"] = TagDefs["!EL"]  = ElementMetaTag;}
        {TagDef AttListMetaTag;
            DefAt(AttListMetaTag, "ELEMENT-NAME", "");
            DefAt(AttListMetaTag, "ATTRIBUTE-LIST", "");
            AttListMetaTag.Handler = &Transceiver::HandleAttListMetaTag;
        TagDefs["!ATTLIST"] = TagDefs["!AT"] = AttListMetaTag;}
        {TagDef EntityMetaTag;
            DefAt(EntityMetaTag, "NAME", "");
            DefAt(EntityMetaTag, "VALUE", "");
            DefAt(EntityMetaTag, "DESC", "");
            EntityMetaTag.Handler = &Transceiver::HandleEntityMetaTag;
        TagDefs["!ENTITY"] = TagDefs["!EN"] = EntityMetaTag;}
        /* VAR not supported yet */
        
        /* ENTITIES */
        
        {EntityDef ent; ent.Value="<"; EntityDefs["lt"]=ent;}
        {EntityDef ent; ent.Value=">"; EntityDefs["gt"]=ent;}
        {EntityDef ent; ent.Value="\""; EntityDefs["quot"]=ent;}
        {EntityDef ent; ent.Value="&"; EntityDefs["amp"]=ent;}
        {EntityDef ent; ent.Value=""; EntityDefs["nbsp"]=ent;}
        
        /* TEXT FORMATTING TAGS (OPEN) */
        {TagDef Tag; Tag.Is_EmptyTag = false; Tag.Is_OpenTag = true;
            Tag.Handler = &Transceiver::HandleBoldTag;
         TagDefs["B"] = TagDefs["STRONG"] = TagDefs["BOLD"] = Tag;}
        {TagDef Tag; Tag.Is_EmptyTag = false; Tag.Is_OpenTag = true;
            Tag.Handler = &Transceiver::HandleItalicTag;
         TagDefs["I"] = TagDefs["ITALIC"] = TagDefs["EM"] = Tag;}
        {TagDef Tag; Tag.Is_EmptyTag = false; Tag.Is_OpenTag = true;
            Tag.Handler = &Transceiver::HandleUnderlineTag;
         TagDefs["U"] = TagDefs["UNDERLINE"] = Tag;}
        {TagDef Tag; Tag.Is_EmptyTag = false; Tag.Is_OpenTag = true;
            Tag.Handler = &Transceiver::HandleStrikeoutTag;
         TagDefs["S"] = TagDefs["STRIKE"] = TagDefs["STRIKEOUT"] = Tag;}
        {TagDef Tag; Tag.Is_EmptyTag = false; Tag.Is_OpenTag = true;
            DefAt(Tag, "FORE", "");
            DefAt(Tag, "BACK", "");
            Tag.Handler = &Transceiver::HandleColorTag;
         TagDefs["C"] = TagDefs["COLOR"] = TagDefs["COLOUR"] = Tag;}
        {TagDef Tag; Tag.Is_EmptyTag = false; Tag.Is_OpenTag = true;
            Tag.Handler = &Transceiver::HandleBrightTag;
         TagDefs["H"] = TagDefs["HIGH"] = Tag;}
        {TagDef Tag; Tag.Is_EmptyTag = false; Tag.Is_OpenTag = true;
            DefAt(Tag, "FACE", "");
            DefAt(Tag, "SIZE", "");
            DefAt(Tag, "COLOR", "");
            DefAt(Tag, "BACK", "");
            Tag.Handler = &Transceiver::HandleFontTag;
         TagDefs["FONT"] = Tag;}

        /* TEXT FORMATTING TAGS (SECURE) */
        {TagDef Tag; Tag.Is_EmptyTag = false;
            Tag.Handler = &Transceiver::HandleNobrTag;
         TagDefs["NOBR"] = Tag;}
        {TagDef Tag; Tag.Is_EmptyTag = false;
            Tag.Handler = &Transceiver::HandleParaTag;
         TagDefs["P"] = Tag;}
        {TagDef Tag; Tag.Is_EmptyTag = true;
            Tag.Handler = &Transceiver::HandleBRTag;
         TagDefs["BR"] = Tag;}
        {TagDef Tag; Tag.Is_EmptyTag = true;
            Tag.Handler = &Transceiver::HandleSBRTag;
         TagDefs["SBR"] = Tag;}

        {TagDef Tag; Tag.Is_EmptyTag = false;
            Tag.Handler = &Transceiver::HandleSupportTag;
         TagDefs["SUPPORT"] = Tag;}
    }

private:
    bool TranslateZcommand(const std::string& param)
    {
        #define NlMode(n) ModeMode=UntilNl,CurrentMode=n##Mode
        #define LkMode(n) ModeMode=ModeLocked,CurrentMode=n##Mode
        #define TgMode(n) ModeMode=UntilTagEnd,CurrentMode=n##Mode
        
        if(param == "2;0") { ClosingTagsMode=false; return true; }
        if(param == "2;1") { ClosingTagsMode=true; return true; }
        
        unsigned val = std::atoi(param.c_str());
        switch(val)
        {
            case 0: NlMode(Open); return true;
            case 1: NlMode(Secure); return true;
            case 2: NlMode(Locked); return true;
            case 3: Formatting.Reset(); NlMode(Open); return true;
            case 4: TgMode(Secure); return true;
            case 5: LkMode(Open); return true;
            case 6: LkMode(Secure); return true;
            case 7: LkMode(Locked); return true;
        }
        return false;
    }
    
    bool TranslateTag(const Tag& tag)
    {
        std::map<std::string/*tagname*/, TagDef>::const_iterator
            i = TagDefs.find(tag.TagName);
        if(i != TagDefs.end())
        {
            if(i->second.Is_EmptyTag && tag.IsClosingTag) return false;
            
            (this->*(i->second.Handler))(i->second, tag);
            
            return true;
        }
        return false;
    }
    
    void DoMacroExpansion(const std::string& macro)
    {
        PendingMacro += macro;
    }
    
    void RedoFormatting()
    {
        /* Translates the current MXP formattings into
         * Xterm-256color compatible escape codes. */ 
        std::vector<unsigned> defs;
        
        static bool was_fore = true;
        static bool was_back = true;
        static bool was_B = false;
        static bool was_U = false;
        
        unsigned forec = 0xC0C0C0; bool foreset=false;
        unsigned backc = 0x000000; bool backset=false;
        
        for(std::vector<FontDef>::const_iterator
            i = Formatting.FontDefs.begin();
            i != Formatting.FontDefs.end();
            ++i)
        {
            if(!i->forecolordefs.empty())
                { foreset=true; forec = TranslateColorNameToRGB(i->forecolordefs); }
            if(!i->backcolordefs.empty())
                { backset=true; backc = TranslateColorNameToRGB(i->backcolordefs); }
        }
        
        #define addm(n) defs.push_back((n))
        
        /* xterm-256color escapes:
          SET:   1=bold 4=underscore 5=blink 7=reverse 8=invisible
          UNSET: 2x  with the same as above
          
          FORE:  SET: 3x {x=0..7}, 38;5;x {x=0..255}
                 UNSET: 39
          BACK:  Same except 4 instead of 3
          
          0 resets colors and flags.
        */
        
        if(Formatting.Bright) { if(foreset) forec |= 0xC0C0C0; }

        if(foreset)
        {
            const unsigned xtermcolor = CreateXtermColor(forec);
            if(xtermcolor < 8)
                addm(30 + xtermcolor);
            else
                { addm(38); addm(5); addm(xtermcolor); }
            was_fore = true;
        }
        else if(was_fore)
            { was_fore=false; addm(39); }

        if(backset)
        {
            const unsigned xtermcolor = CreateXtermColor(backc);
            if(xtermcolor < 8)
                addm(40 + xtermcolor);
            else
                { addm(48); addm(5); addm(xtermcolor); }
            was_back=true;
        }
        else if(was_back)
            { was_back=false; addm(49); }

        if(Formatting.B) { if(!was_B) { was_B=true; addm(1); } } else if(was_B) { was_B=false; addm(22); }
        if(Formatting.I) {}
        if(Formatting.U) { if(!was_U) { was_U=true; addm(4); } } else if(was_U) { was_U=false; addm(24); }
        if(Formatting.S) {}
        
        #undef addm

        std::string result = "\33[";
        for(unsigned a=0; a<defs.size(); ++a)
        {
            char Buf[32];
            sprintf(Buf, "%s%u", a?";":"", defs[a]);
            result += Buf;
        }
        result += 'm';
        PendingResult += result;
    }

    void HandleChar(const unsigned char input_ch)
    {
        scrBegin;
        #define NextCh() scrReturnV
        for(;;)
        {
            //fprintf(stderr, "received ch: '%c'\n", input_ch);
            
            if(input_ch == IAC) // telnet escape
            {
                static std::string telnetesc; telnetesc = "";
                telnetesc += input_ch;
                
                NextCh();
                if(input_ch == WILL)
                {
                    telnetesc += input_ch;
                    NextCh();
                    if(input_ch == 91) // MXP
                    {
                        PendingResult += (char)IAC;
                        PendingResult += (char)DO;
                        PendingResult += (char)91;
                        continue;
                    }
                }
                PendingResult += telnetesc;
                goto Literal;
            }
            if(CurrentMode != LockedMode)
            {
                if(input_ch == '\33')
                {
                    NextCh();
                    if(input_ch != '[') { PendingResult += '\33'; goto Literal; }
                    static std::string params; params="";
                NextParamCh:
                    NextCh();
                    if((input_ch >= '0' && input_ch <= '9') || input_ch==';')
                        { params += input_ch; goto NextParamCh; }
                    if(input_ch == 'z')
                        { if(!TranslateZcommand(params))
                            goto LiteralEscCmd;
                        }
                    else
                    { LiteralEscCmd:
                        PendingResult += "\33[";
                        PendingResult += params;
                        goto Literal;
                    }
                }
                else if(input_ch == '<') // tag
                {
                    static std::string tagbuf; tagbuf = "";
                    #define TagGetCh() { tagbuf += input_ch; NextCh(); }
                    static std::string tagname; tagname = "";
                    static bool is_closing; is_closing=false;
                    static std::string attrlist; attrlist="";
                    for(;;)
                    {
                        TagGetCh();
                        //fprintf(stderr, "Tagname processing: '%c'\n", input_ch);
                        if(tagname.empty() && input_ch == '/') { is_closing=true; continue; }
                        if(std::isspace(input_ch) || input_ch == '>' || input_ch == '/')
                        {
                            if(tagname.empty()) continue;
                            break;
                        }
                        if(tagname.empty() && input_ch == '>'
                        || tagname.size() > 48
                        || !is_valid_tagname_char(input_ch))
                            { PendingResult += tagbuf; goto Literal; }
                        tagname += input_ch;
                    }
                    while(std::isspace(input_ch)) { TagGetCh(); }
                    static char quotemode; quotemode=0;
                    for(;;)
                    {
                        //fprintf(stderr, "Tagattr processing: '%c'\n", input_ch);
                        if(!quotemode && input_ch == '/') { TagGetCh(); continue; }
                        if(!quotemode && input_ch == '>') break;
                        attrlist += input_ch;
                        if(quotemode && input_ch==quotemode) { quotemode=false; }
                        else if(!quotemode && (input_ch=='"' || input_ch=='\'')) { quotemode=input_ch; }
                        else {}
                        TagGetCh();
                    }
                    
                    //fprintf(stderr, "name(%s)attr(%s)\n",
                    //    tagname.c_str(), attrlist.c_str());
                    
                    Tag tag;
                    tag.TagName      = UCase(tagname);
                    tag.IsClosingTag = is_closing;
                    if(ClosingTagsMode) tag.IsClosingTag = true; // force closing
                    tag.Params = ParseAttrList(attrlist, false);
                    if(!TranslateTag(tag))
                    {
                        PendingResult += tagbuf;
                        goto Literal;
                    }
                    #undef TagGetCh
                }
                else if(input_ch == '&') // entity
                {
                    static std::string entity; entity = "";
                    NextCh();
                    if(input_ch == '#')
                    {
                        static unsigned charvalue; charvalue = 0;
                        static unsigned base; base = 10;
                        entity += '#';
                        NextCh(); if(input_ch == 'x' || input_ch == 'X')
                                   { entity += input_ch; base=16; NextCh(); }
                        for(;;)
                        {
                            #define chr(min,max, sub,add) \
                                if(input_ch >= min && input_ch <= max) \
                                { charvalue = charvalue*base + (input_ch - sub + add); \
                                  entity += input_ch; \
                                  NextCh(); continue; }

                            chr('0','9', '0',0);
                            if(base == 16)
                            { chr('A','F', 'A',10);
                              chr('a','f', 'a',10); }
                            if(input_ch == ';') break;
                        FailedEntity:
                            PendingResult += "&";
                            PendingResult += entity;
                            goto Literal;
                        }
                        PendingResult += (char)charvalue;
                    }
                    else
                    {
                        while(input_ch != ';' && entity.size() < 16)
                            { entity += input_ch; NextCh(); }
                        if(input_ch != ';') goto FailedEntity;
                        std::map<std::string, EntityDef>::const_iterator
                            i = EntityDefs.find(entity);
                        if(i == EntityDefs.end()) goto FailedEntity;
                        
                        if(i->second.Is_Recursive)
                            DoMacroExpansion(i->second.Value);
                        else
                            PendingResult += i->second.Value;
                    }
                }
                else goto Literal;
            }
            else
            {
            Literal:
                PendingResult += input_ch;
                if(ModeMode == UntilTagEnd) goto ResetDefaultMode;
                if(input_ch == '\n' && ModeMode == UntilNl)
                {
                ResetDefaultMode:
                    CurrentMode = DefaultMode;
                    ModeMode    = ModeLocked;
                }
            }
            NextCh();
        }
        scrFinishV;
    }
public:
    void ProcessBuf(const std::string& buf)
    {
        for(unsigned a=0; a<buf.size(); ++a)
        {
            HandleChar(buf[a]);
            
            while(!PendingMacro.empty())
            {
                const std::string macro = PendingMacro;
                PendingMacro = "";
                for(unsigned b=0; b<macro.size(); ++b)
                    HandleChar(macro[b]);
            }
        }
    }
public:
    Transceiver() :
        CurrentMode(DefaultMode),
        ModeMode(ModeLocked),
        ClosingTagsMode(false)
    {
        InitDefaultTags();
    }
    
    ssize_t Handle(ssize_t(*orig_recv)(int,void*,size_t,int),
                   int s,void*buf,size_t len,int flags)
    {
    Redo:
        if(len > 0 && !PendingResult.empty())
        {
            ssize_t take = std::min(len, PendingResult.size());
            std::memcpy(buf, PendingResult.data(), take);
            PendingResult.erase(0, take);
            return take;
        }
    
        ssize_t res = orig_recv(s,buf,len,flags);
        if(res <= 0) return res;

        const char* charbuf = (const char*)buf;
        std::string Buf( charbuf, charbuf + res);
        //PendingResult += Buf;
        ProcessBuf(Buf);
        goto Redo;
    }
    
    const std::string& GetResult() const { return PendingResult; }
};

extern "C" {

ssize_t recv(int s, void *buf, size_t len, int flags)
{
    /* Find the real fork function */
    typedef ssize_t (*old_t)(int,void*,size_t,int);
    static old_t old = (old_t)caller("recv");
    
    static class Transceiver Transceiver;
    return Transceiver.Handle(old, s,buf,len,flags);
}

}

/*
int main(void)
{
    Transceiver t;
    t.ProcessBuf("<C #000000> Pippelip </C> x");
    fprintf(stderr, "Got: '%s'\n", t.GetResult().c_str());
}*/
